package alterrs.jbct.utility;

import static org.objectweb.asm.Opcodes.ACC_INTERFACE;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.lang.reflect.Modifier;
import java.net.JarURLConnection;
import java.net.URL;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;
import java.util.jar.JarOutputStream;

import org.objectweb.asm.ClassReader;
import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.Opcodes;
import org.objectweb.asm.commons.JSRInlinerAdapter;
import org.objectweb.asm.commons.LocalVariablesSorter;
import org.objectweb.asm.tree.AbstractInsnNode;
import org.objectweb.asm.tree.ClassNode;
import org.objectweb.asm.tree.FieldNode;
import org.objectweb.asm.tree.MethodInsnNode;
import org.objectweb.asm.tree.MethodNode;

import alterrs.commons.Attributes;

/**
 * @author Lazaro
 */
public class ClassCollection {
	private URL jarURL;

	private Map<String, ClassNode> classes = null;
	private Attributes attributes = new Attributes();
	private Refactorer refactorer = new Refactorer(this);

	private Map<ID, ClassNode> cachedExternalClasses = new HashMap<ID, ClassNode>();

	public ClassCollection(URL jarURL) {
		this.jarURL = jarURL;
	}

	public void load() throws IOException {
		classes = new HashMap<String, ClassNode>();

		JarFile jar = ((JarURLConnection) jarURL.openConnection()).getJarFile();
		Enumeration<JarEntry> entries = jar.entries();
		if (!entries.hasMoreElements()) {
			return;
		}
		JarEntry entry;
		while (entries.hasMoreElements()) {
			entry = entries.nextElement();

			if (!entry.getName().endsWith(".class")) {
				continue;
			}

			ClassNode c = new ClassNode();
			ClassReader cr = new ClassReader(jar.getInputStream(jar
					.getEntry(entry.getName())));
			cr.accept(c, ClassReader.SKIP_FRAMES);

			classes.put(c.name, c);
		}
	}

	public void transform() {
		refactorer.run();
		refactorer.clear();
	}

	public void save(File dst) throws IOException {
		ClassCollection cl = this;
		
		JarOutputStream jos = new JarOutputStream(new FileOutputStream(dst));
		for (ClassNode c : classes.values()) {
			ClassWriter cw = new ClassWriter(ClassWriter.COMPUTE_FRAMES | ClassWriter.COMPUTE_MAXS) {
				// this method in ClassWriter uses the systemclassloader as
				// a stream location to load the super class, however, most of
				// the time the class is loaded/read and parsed by us so it
				// isn't defined in the system classloader. in certain cases
				// we may not even want it to be loaded/resolved and we can
				// bypass this by implementing the hierarchy scanning algorithm
				// with ClassNodes rather than Classes.
			    @Override
				protected String getCommonSuperClass(String type1, String type2) {
			    	ClassNode ccn = cl.getClass(type1);
			    	ClassNode dcn = cl.getClass(type2);
			    	
			    	if(ccn == null) {
			    		throw new RuntimeException("ccn is null," + type1);
			    	}
			    	
			    	if(dcn == null) {
			    		throw new RuntimeException("dcn is null," + type2);
			    	}
			    	
			    	List<ClassNode> c = cl.getSupers(ccn);
			    	List<ClassNode> d = cl.getSupers(dcn);
			        
			        if(c.contains(dcn))
			        	return type1;
			        
			        if(d.contains(ccn))
			        	return type2;
			        
			        if(Modifier.isInterface(ccn.access) || Modifier.isInterface(dcn.access)) {
			        	// enums as well?
			        	return "java/lang/Object";
			        } else {
			        	do {
			        		ClassNode nccn = cl.getClass(ccn.superName);
			        		if(nccn == null)
			        			break;
			        		ccn = nccn;
			        		c = cl.getSupers(ccn);
			        	} while(!c.contains(dcn));
			        	return ccn.name;
			        }
			    }
			};
			c.accept(cw);
			JarEntry je = new JarEntry(c.name + ".class");
			jos.putNextEntry(je);
			jos.write(cw.toByteArray());
			jos.closeEntry();
		}
		jos.close();
	}

	@SuppressWarnings("unchecked")
	public void inlineJSR() {
		for (String key : classes.keySet()) {
			ClassNode n = classes.get(key);
			
			//Create new class node
			ClassNode node = new ClassNode();
			//Populate visitors
			n.accept(node);
			//Delete copied methods
			node.methods.clear();
			for (Object mo : n.methods) {
				//Cast to method node
				MethodNode mn = (MethodNode) mo;
				//Create new method node
				MethodNode method = new MethodNode(Opcodes.ASM5, mn.access, mn.name, mn.desc, mn.signature, ((List<String>) mn.exceptions).toArray(new String[mn.exceptions.size()]));
				//Create new jsr inliner which emits to the new method
				JSRInlinerAdapter adapter = new JSRInlinerAdapter(method, mn.access, mn.name, mn.desc, mn.signature, ((List<String>) mn.exceptions).toArray(new String[mn.exceptions.size()]));
				//Populate visitors
				mn.accept(adapter);
				//Append method to new class
				node.methods.add(method);
			}
			//Append class to table
			classes.put(key, node);
		}
	}
	
	@SuppressWarnings("unchecked")
	public void sortVars() {
		for (String key : classes.keySet()) {
			ClassNode n = classes.get(key);
			
			//Create new class node
			ClassNode node = new ClassNode();
			//Populate visitors
			n.accept(node);
			//Delete copied methods
			node.methods.clear();
			for (Object mo : n.methods) {
				//Cast to method node
				MethodNode mn = (MethodNode) mo;
				//Create new method node
				MethodNode method = new MethodNode(Opcodes.ASM5, mn.access, mn.name, mn.desc, mn.signature, ((List<String>) mn.exceptions).toArray(new String[mn.exceptions.size()]));
				//Create new locals sorter which emits to the new method
				LocalVariablesSorter sorter = new LocalVariablesSorter(mn.access, mn.desc, method);
				//Populate visitors
				mn.accept(sorter);
				//Append method to new class
				node.methods.add(method);
			}
			//Append class to table
			classes.put(key, node);
		}
	}
	
	public Map<String, ClassNode> getClasses() {
		return classes;
	}

	public Attributes getAttributes() {
		return attributes;
	}

	public Refactorer getRefactorer() {
		return refactorer;
	}

	public FieldNode field(FieldID id) {
		ClassNode c = classes.get(id.owner);
		if (c != null) {
			for (Object fo : c.fields) {
				FieldNode f = (FieldNode) fo;
				if (f.name.equals(id.name) && f.desc.equals(id.desc)) {
					return f;
				}
			}
		}
		return null;
	}

	public MethodNode method(MethodID id) {
		ClassNode c = classes.get(id.owner);
		if (c != null) {
			for (Object mo : c.methods) {
				MethodNode m = (MethodNode) mo;
				if (m.name.equals(id.name) && m.desc.equals(id.desc)) {
					return m;
				}
			}
		}
		c = getClass(id.owner);
		if (c != null) {
			for (Object mo : c.methods) {
				MethodNode m = (MethodNode) mo;
				if (m.name.equals(id.name) && m.desc.equals(id.desc)) {
					return m;
				}
			}
		}
		return null;
	}

	public List<ClassNode> getInterfaces() {
		List<ClassNode> list = new ArrayList<ClassNode>();
		for (ClassNode c : getClasses().values()) {
			if ((c.access & ACC_INTERFACE) != 0) {
				list.add(c);
			}
		}
		return list;
	}

	public boolean isMethodUsed(MethodID id) {
		ClassNode c = getClass(id.owner);		

		if (c == null) {
			System.err.println("ClassNode is null! owner="+id.owner);
			return false;
		}

		for (ClassNode c2 : getInterfaces(c)) { // loop interfaces -- so that we dont remove unused inherited methods
			if (c2 != c) {
				for (Object mo : c2.methods) {
					MethodNode m2 = (MethodNode) mo;
					if (m2.name.equals(id.name) && m2.desc.equals(id.desc)) {
						return true;
					}
				}
			}
		}

		for (ClassNode c2 : getHierarchy(c)) { // loop super classes
			if (c2 != c) {
				for (Object mo : c2.methods) { // loop through the method of each class
					MethodNode m2 = (MethodNode) mo;
					if (m2.name.equals(id.name) && m2.desc.equals(id.desc)) {
						return true;
					}
				}
			}
		}
		
		for(ClassNode n2 : classes.values()) { //loop through all classes
			if (n2 != c) { // class cant equal analyzing class
				for (Object mo : n2.methods) { // loop through the method of each class
					MethodNode m2 = (MethodNode) mo;
					if (m2.name.equals(id.name) && m2.desc.equals(id.desc)) {
						return true;
					}
				}
			}
		}
		
		if (((c.access & ACC_INTERFACE) != 0)) {
			for (ClassNode c2 : getInterfaces()) {
				if (c2 != c) {
					for (Object mo : c2.methods) {
						MethodNode m2 = (MethodNode) mo;
						if (m2.name.equals(id.name) && m2.desc.equals(id.desc)) {
							return true;
						}
					}
				}
			}
		}

		/**
		 * Check everything else above before this, because this will take the longest..
		 */
		for(ClassNode n2 : classes.values()) { // loop classes -- this is where it gets busy
			for (Object mo : n2.methods) { // loop through the method of each class
				MethodNode m2 = (MethodNode) mo;
				AbstractInsnNode[] insns = m2.instructions.toArray();
				for (AbstractInsnNode insn : insns) { // loop through each instruction to see if this method is called.
					if (insn instanceof MethodInsnNode) {
						MethodInsnNode node = ((MethodInsnNode) insn);
						if (node.name.equals(id.name) && node.desc.equals(id.desc))
							return true;
					}
				}
			}
		}

		return false;
	}
	
	public ClassNode[] getDeclorators(MethodID id) {
		ClassNode c = getClass(id.owner);
		
		List<ClassNode> list = new ArrayList<ClassNode>();		

		if (c == null) {
			System.err.println("Class is null in 'getDeclorators', " + id.owner);
			return list.toArray(new ClassNode[0]);
		}

		for (ClassNode c2 : getInterfaces(c)) {
			if (c2 != c) {
				for (Object mo : c2.methods) {
					MethodNode m2 = (MethodNode) mo;
					if (m2.name.equals(id.name) && m2.desc.equals(id.desc)) {
						list.add(c2);
					}
				}
			}
		}

		if (list.isEmpty()) {
			hierarchyLoop: for (ClassNode c2 : getHierarchy(c)) {
				if (c2 != c) {
					for (Object mo : c2.methods) {
						MethodNode m2 = (MethodNode) mo;
						if (m2.name.equals(id.name) && m2.desc.equals(id.desc)) {
							list.add(c2);
							break hierarchyLoop;
						}
					}
				}
			}
		}

		if (list.isEmpty() && ((c.access & ACC_INTERFACE) != 0)) {
			for (ClassNode c2 : getInterfaces()) {
				if (c2 != c) {
					for (Object mo : c2.methods) {
						MethodNode m2 = (MethodNode) mo;
						if (m2.name.equals(id.name) && m2.desc.equals(id.desc)) {
							list.add(c2);
						}
					}
				}
			}
		}

		if (list.isEmpty() || ((c.access & ACC_INTERFACE) != 0)) {
			list.add(c);
		}

		return list.toArray(new ClassNode[0]);
	}

	public ClassNode[] getInheritors(MethodID id) {
		ClassNode c = getClass(id.owner);

		List<ClassNode> inheritors = new LinkedList<ClassNode>();
		classLoop: for (ClassNode c2 : getClasses().values()) {
			ClassNode[] inherited = (c.access & ACC_INTERFACE) == 0 ? getHierarchy(c2)
					: getInterfaces(c2);
			int index = ArrayUtil.searchArray(inherited, c);
			if (index >= 0) {
				for (Object mo : c2.methods) {
					MethodNode m2 = (MethodNode) mo;
					if (m2.name.equals(id.name) && m2.desc.equals(id.desc)
							&& !inheritors.contains(c2)) {
						inheritors.add(c2);
						continue classLoop;
					}
				}
			}
		}
		return inheritors.toArray(new ClassNode[0]);
	}

	public ClassNode[] getHierarchy(ClassNode c) {
		LinkedList<ClassNode> hierarchy = new LinkedList<ClassNode>();

		ClassNode c2 = c;
		while (c2 != null) {
			hierarchy.addFirst(c2);

			if (c2.superName == null) {
				break;
			}

			c2 = getClass(c2.superName);
		}

		return hierarchy.toArray(new ClassNode[0]);
	}
	
	public List<ClassNode> getSupers(ClassNode c) {
		LinkedList<ClassNode> hierarchy = new LinkedList<ClassNode>();

		ClassNode c2 = c;
		while (c2 != null) {
			hierarchy.addFirst(c2);

			if (c2.superName == null) {
				break;
			}

			c2 = getClass(c2.superName);
		}

		return hierarchy;
	}

	public ClassNode[] getInterfaces(ClassNode c) {
		LinkedList<ClassNode> interfaces = new LinkedList<ClassNode>();

		ClassNode[] hierarchy = getHierarchy(c);

		for (ClassNode c2 : hierarchy) {
			if ((c2.access & ACC_INTERFACE) != 0) {
				interfaces.add(c2);
			}

			for (Object c3o : c2.interfaces) {
				ClassNode c3 = getClass((String) c3o);
				if (c3 == null)
					throw new RuntimeException("Class is null in 'getInterfaces', " + (String) c3o);
				
				interfaces.addFirst(c3);

				for (ClassNode c4 : getInterfaces(c3)) {
					interfaces.addFirst(c4);
				}
			}
		}

		return interfaces.toArray(new ClassNode[0]);
	}


	public ClassNode getClass(String name) {
		if (name == null) {
			return null;
		}
		
		if (name.contains("["))
			name = "java/lang/Object";

		ClassNode c = getClasses().get(name);
		if (c == null) {
			c = cachedExternalClasses.get(new ID(name));

			if (c == null) {
				c = new ClassNode();
				ClassReader cr;
				try {
					cr = new ClassReader(name);
				} catch (Exception e) {
					return null;
				}
				cr.accept(c, 0);

				cachedExternalClasses.put(new ID(name), c);
			}
		}
		return c;
	}

	public int getLevel(ClassNode c) {
		if ((c.access & ACC_INTERFACE) != 0) {
			return 0;
		}

		return getHierarchy(c).length - 1;
	}

	public boolean isExternal(ID id) {
		return cachedExternalClasses.containsKey(id);
	}
}
