package com.method.rscd;

import com.method.rscd.cache.ReferenceTable;
import com.method.rscd.cache.store.BufferedDiskFile;
import com.method.rscd.cache.store.CacheStore;
import com.method.rscd.cache.store.DiskFile;
import com.method.rscd.net.FileRequest;
import com.method.rscd.net.HttpWorker;
import com.method.rscd.net.Js5Worker;
import com.method.rscd.util.Crawler;
import com.method.rscd.util.Whirlpool;

import java.io.File;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.*;
import java.util.zip.CRC32;

/**
 * Coordinates downloading and saving of the cache from the server.
 */
public class CacheDownloader {

    public static final boolean DOWNLOAD_MUSIC = true;

    private static final int MAJOR_VERSION = 912;
    private static final int MINOR_VERSION = 1;
    private static final String HOST = "content.runescape.com";
    private static final int WORLD = 31;
    private static final int LANGUAGE = 0; // English language by default

    private static final int MUSIC_INDEX = 40;
    private static final int PROGRESS_TICKS = 20;

    private Js5Worker js5Worker;
    private HttpWorker httpWorker;
    private ReferenceTable versionTable;
    private ReferenceTable[] tables;
    private ReferenceTable[] oldTables;
    private CacheStore reference;
    private CacheStore[] stores;

    /**
     * Entry point for the application. This method creates a new CacheDownloader
     * instance and runs it.
     *
     * @param args Arguments for the program
     */
    public static void main(String[] args) {
        CacheDownloader downloader = new CacheDownloader();
        downloader.run();
    }

    /**
     * Creates a new CacheDownloader object.
     */
    public CacheDownloader() {
        js5Worker = new Js5Worker();
        httpWorker = new HttpWorker(HOST);
    }

    /**
     * Initiates the connection to the server and downloads the cache.
     */
    public void run() {
        connect();
        downloadVersionTable();
        initCacheIndices(versionTable.getEntryCount());
        initOldTables();
        downloadNewTables();
        update();

        httpWorker.shutdown();
    }

    /**
     * Connects to the server, retrying as needed if the version is incorrect.
     */
    private void connect() {
        String java = Crawler
                .extractParameter(Crawler.downloadPage("http://world" + WORLD + ".runescape.com/g=runescape/,j1"), 32,
                        true);
        String nxt = Crawler.extractParameter(Crawler.downloadPage(
                "http://world" + WORLD + ".runescape.com/k=5/jav_config.ws?userFlow=1337&binaryType=2"), 32, false);
       /* String key = nxt;
        if (key == null) {
            System.err.println("Couldn't find valid handshake key.");
            System.exit(1);
        }*/
        String[] keys = {nxt};
        for (String key : keys) {
            js5Worker.connect(HOST, MAJOR_VERSION, MINOR_VERSION, key, LANGUAGE);
            while (js5Worker.getState() != Js5Worker.State.CONNECTED) {
                js5Worker.process();
            }
            System.out.println("Successful connection");
        }
    }

    /**
     * Downloads the version table from the server.
     */
    private void downloadVersionTable() {
        FileRequest mainRequest = js5Worker.request(255, 255);
        while (!mainRequest.isComplete()) {
            js5Worker.process();
        }
        versionTable = new ReferenceTable(mainRequest.getBuffer());
    }

    /**
     * Reads the existing reference table data from the cache.
     */
    private void initOldTables() {
        oldTables = new ReferenceTable[reference.getFileCount()];
        for (int i = 0; i < oldTables.length; i++) {
            byte[] raw = reference.load(i);
            if (raw != null) {
                ByteBuffer data = ByteBuffer.wrap(raw);
                oldTables[i] = new ReferenceTable(i, data, null);
            }
        }
    }

    /**
     * Finds reference tables that need updating.
     *
     * @return A list containing the reference tables that need updating.
     */
    private List<Integer> findTableChanges() {
        List<Integer> changes = new ArrayList<>();
        ReferenceTable.Entry[] entries = versionTable.getEntries();
        for (int i = 0; i < versionTable.getEntryCount(); i++) {
            ReferenceTable.Entry entry = entries[i];
            if (entry.getCRC() == 0 && entry.getVersion() == 0) {
                continue;
            }
            if (i >= oldTables.length) {
                changes.add(i);
            } else {
                ReferenceTable table = oldTables[i];
                if (table != null) {
                    int crc = table.getCRC(), version = table.getVersion();
                    if (crc != entry.getCRC() || version != entry.getVersion()) {
                        changes.add(i);
                    }
                } else {
                    changes.add(i);
                }
            }
        }

        return changes;
    }

    /**
     * Downloads the required reference tables.
     */
    private void downloadNewTables() {
        List<Integer> changes = findTableChanges();
        Queue<FileRequest> requests = new LinkedList<>();
        tables = new ReferenceTable[versionTable.getEntryCount()];
        for (Integer change : changes) {
            requests.offer(js5Worker.request(255, change));
        }

        while (requests.size() > 0) {
            js5Worker.process();
            for (Iterator<FileRequest> iter = requests.iterator(); iter.hasNext(); ) {
                FileRequest request = iter.next();
                if (request.isComplete()) {
                    int file = request.getFile();
                    ByteBuffer buffer = request.getBuffer();
                    tables[file] = new ReferenceTable(file, buffer, versionTable);
                    buffer.position(0);
                    byte[] data = new byte[buffer.remaining()];
                    buffer.get(data);
                    reference.store(file, data, data.length);
                    iter.remove();
                }
            }
        }
    }

    /**
     * Finds the files that need updating within a particular cache index.
     *
     * @param index The index of the cache to look through
     * @return A list containing the files that need to be updated.
     */
    private List<Integer> findFileChanges(int index) {
        if (index == 40 && !DOWNLOAD_MUSIC) {
            return Collections.emptyList();
        }
        String message = "Checking index " + index;

        List<Integer> changes = new ArrayList<>();
        ReferenceTable.Entry tableEntry = versionTable.getEntries()[index];
        if (tableEntry.getCRC() == 0 && tableEntry.getVersion() == 0) {
            return null;
        }
        if (index >= stores.length && tables[index] != null) {
            ReferenceTable.Entry[] entries = tables[index].getEntries();
            int count = entries.length;
            printProgress(0, count, message);
            for (int i = 0; i < count; i++) {
                changes.add(entries[i].getIndex());
            }
            printProgress(count, count, message);
        } else {
            CRC32 crc32 = new CRC32();
            ReferenceTable table = tables[index] != null || index >= oldTables.length ? tables[index] : oldTables[index];
            if (table == null) {
                return null;
            }
            ReferenceTable.Entry[] entries = table.getEntries();
            int entryCount = table.getEntryCount();
            int interval = Math.max(entryCount / 100, 1);
            for (int i = 0; i < entryCount; i++) {
                ReferenceTable.Entry entry = entries[i];
                int entryIndex = entry.getIndex();
                byte[] data = stores[index].load(entryIndex);
                if (data == null) {
                    changes.add(entryIndex);
                } else {
                    ByteBuffer buffer = ByteBuffer.wrap(data);
                    crc32.update(buffer.array(), 0, buffer.capacity() - 2);
                    int crc = (int) crc32.getValue();
                    crc32.reset();
                    if (crc != entry.getCRC()) {
                        changes.add(entryIndex);
                        continue;
                    }
                    buffer.position(buffer.capacity() - 2);
                    int version = buffer.getShort() & 0xffff;
                    if (version != (entry.getVersion() & 0xffff)) {
                        changes.add(entryIndex);
                    }
                }

                if (i % interval == 0) {
                    printProgress(i, entryCount, message);
                }
            }

            printProgress(entryCount, entryCount, message);
        }
        return changes;

    }

    /**
     * Downloads all of the files from each index in the cache.
     */
    private void update() {
        for (int i = 0; i < versionTable.getEntryCount(); i++) {
            String message = "Updating index " + i;

            List<Integer> changes = shouldIgnore(i) ? null : findFileChanges(i);
            if (changes == null || changes.size() == 0) {
                continue;
            }

            ReferenceTable table = tables[i] != null ? tables[i] : oldTables[i];
            CRC32 crc = new CRC32();

            Queue<FileRequest> requests = new LinkedList<>();
            int total = changes.size();

            for (int j = 0; j < total; j++) {
                int file = changes.get(j);
                FileRequest request = null;
                if (shouldIgnore(i)) {
                    request = null;
                } else if (i == MUSIC_INDEX) {
                    if (DOWNLOAD_MUSIC) {
                        request = httpWorker
                                .request(i, file, table.getEntry(file).getCRC(), table.getEntry(file).getVersion());
                    }
                } else {
                    request = js5Worker.request(i, file);
                }
                if (request != null) {
                    requests.offer(request);
                }
            }

            printProgress(0, total, message);

            int count = 0;
            while (requests.size() > 0) {
                if (i != MUSIC_INDEX) {
                    js5Worker.process();
                }

                for (Iterator<FileRequest> iter = requests.iterator(); iter.hasNext(); ) {
                    FileRequest request = iter.next();
                    if (request.isComplete()) {
                        int file = request.getFile();
                        ByteBuffer buffer = request.getBuffer();
                        ReferenceTable.Entry entry = table.getEntry(file);

                        crc.update(buffer.array(), 0, buffer.limit());
                        if (entry.getCRC() != (int) crc.getValue()) {
                            throw new RuntimeException(
                                    "CRC mismatch " + i + "," + file + "," + entry.getCRC() + " - " + (int) crc
                                            .getValue());
                        }
                        crc.reset();

                        byte[] entryDigest = entry.getDigest();
                        if (entryDigest != null) {
                            byte[] digest = Whirlpool.whirlpool(buffer.array(), 0, buffer.limit());
                            for (int j = 0; j < 64; j++) {
                                if (digest[j] != entryDigest[j]) {
                                    throw new RuntimeException("Digest mismatch " + i + "," + file);
                                }
                            }
                        }

                        int version = entry.getVersion();
                        buffer.position(buffer.limit()).limit(buffer.capacity());
                        buffer.put((byte) (version >>> 8));
                        buffer.put((byte) version);
                        buffer.flip();
                        byte[] data = new byte[buffer.remaining()];
                        buffer.get(data);

                        stores[i].store(file, data, data.length);
                        iter.remove();

                        count++;
                        printProgress(count, total, message);
                    }
                }
            }
        }
    }

    private static boolean shouldIgnore(int id) {
        switch (id) {
            case 14:
            case 40:
                return true;
            default:
                return false;
        }
    }

    /**
     * Initializes the cache indices.
     *
     * @param count The number of indices
     */
    private void initCacheIndices(int count) {
        try {
            String path = "./data/cache/";
            reference = new CacheStore(255,
                    new BufferedDiskFile(new DiskFile(new File(path + "main_file_cache.dat255"), "rw", 32_221_225_472L),
                            5200, 0),
                    new BufferedDiskFile(new DiskFile(new File(path + "main_file_cache.idx255"), "rw", 32_221_225_472L),
                            6000, 0));
            stores = new CacheStore[count];
            for (int i = 0; i < count; i++) {
                stores[i] = new CacheStore(i, new BufferedDiskFile(
                        new DiskFile(new File(path + "main_file_cache.dat" + i), "rw", 32_221_225_472L), 5200, 0),
                        new BufferedDiskFile(
                                new DiskFile(new File(path + "main_file_cache.idx" + i), "rw", 32_221_225_472L), 6000,
                                0));
            }
        } catch (IOException ioex) {
            ioex.printStackTrace();
        }
    }

    private static void printProgress(int current, int total, String message) {
        StringBuilder line = new StringBuilder();
        line.append('\r').append(message).append(" [");

        int ticks;
        int percent;
        if (current == total) {
            ticks = PROGRESS_TICKS;
            percent = 100;
        } else {
            ticks = current * PROGRESS_TICKS / total;
            percent = (int) (current * 100.0f / total);
        }

        for (int i = 0; i < ticks; i++) {
            line.append('=');
        }

        for (int i = 0; i < PROGRESS_TICKS - ticks; i++) {
            line.append(' ');
        }

        line.append("] ").append(percent).append("% ").append(current).append(" / ").append(total);
        if (current == total) {
            System.out.println(line.toString());
        } else {
            System.out.print(line.toString());
        }
    }

}
