package com.method.rscd.cache.store;

import java.io.EOFException;
import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;

/**
 * Created at: Dec 12, 2016 11:39:30 PM
 * 
 * @author Walied-Yassen A.K.A Cody
 */
public class DiskFile {

	/**
	 * The backend class.
	 */
	private RandomAccessFile file;

	/**
	 * The maximum size of this file.
	 */
	private long maxSize;

	/**
	 * Current offset we writing/reading at.
	 */
	private long offset;

	/**
	 * 
	 * Construct a new {@link DiskFile} object instance.
	 * 
	 * @param file
	 *            the backend file.
	 * @param mode
	 *            the file mode ("r","w","rw")
	 * @param maxSize
	 *            the maximum size of this file.
	 * @throws IOException
	 *             if anything occurs.
	 */
	public DiskFile(File file, String mode, long maxSize) throws IOException {
		if (maxSize == -1L) {
			maxSize = Long.MAX_VALUE;
		}
		if (file.length() > maxSize) {
			file.delete();
		}
		this.file = new RandomAccessFile(file, mode);
		this.maxSize = maxSize;
		this.offset = 0L;
		int first = this.file.read();
		if (first != -1 && !mode.equals("r")) {
			this.file.seek(0L);
			this.file.write(first);
		}
		this.file.seek(0L);
	}

	/**
	 * Read an specified amount of bytes into an specified array, starting at specified offset.
	 * 
	 * @param data
	 *            the array to read into.
	 * @param off
	 *            the starting offset of the array.
	 * @param len
	 *            the amount of bytes to read.
	 * @return the number of bytes read.
	 * @throws IOException
	 *             if anything occurs
	 */
	public int read(byte[] data, int off, int len) throws IOException {
		int numRead = file.read(data, off, len);
		if (numRead > 0) {
			offset += numRead;
		}
		return numRead;
	}

	/**
	 * Writes specified length of data starting from specified offset.
	 * 
	 * @param data
	 *            the data to write.
	 * @param off
	 *            the offset to start writing at.
	 * @param len
	 *            the amount of data to write.
	 * @throws IOException
	 *             if anything occurs.
	 */
	public void write(byte[] data, int off, int len) throws IOException {
		if (offset + len > maxSize) {
			file.seek(maxSize);
			file.write(1);
			throw new EOFException("offset > maxSize");
		}
		file.write(data, off, len);
		offset += len;
	}

	/**
	 * Set the current read/write offset.
	 * 
	 * @param offset
	 *            the offset to set.
	 * @throws IOException
	 *             if anything occurs.
	 */
	public void seek(long offset) throws IOException {
		file.seek(offset);
		this.offset = offset;
	}

	/**
	 * Get the current size of the file on the disk.
	 * 
	 * @return the size.
	 * @throws IOException
	 *             if anything occurs.
	 */
	public long getSize() throws IOException {
		return file.length();
	}

	/**
	 * Close the file and stop the I/O for this file.
	 * 
	 * @throws IOException
	 *             if anything occurs.
	 */
	public void close() throws IOException {
		if (file != null) {
			file.close();
			file = null;
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#finalize()
	 */
	@Override
	protected void finalize() throws Throwable {
		if (null != this.file) {
			close();
		}
	}

}
