package rs.emulate.lynx;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.net.URL;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.security.GeneralSecurityException;
import java.util.Map;
import java.util.Map.Entry;
import java.util.jar.JarOutputStream;
import java.util.zip.ZipEntry;

import rs.emulate.lynx.net.ClientVersionWorker;
import rs.emulate.lynx.net.ConfigCrawler;
import rs.emulate.lynx.net.Crawler;
import rs.emulate.lynx.net.Js5Constants;

/**
 * Retrieves and decrypts the Runescape game client.
 * 
 * @author Major
 */
public final class Lynx {

	/**
	 * Indicates whether to identify the current client version, or to use the current date instead.
	 */
	private final boolean identifyVersion;

	/**
	 * Creates the new lynx object.
	 * 
	 * @param identifyVersion Indicates whether or not to identify the current client version.
	 * @param local 
	 */
	public Lynx(boolean identifyVersion) {
		this.identifyVersion = identifyVersion;
	}

	/**
	 * Downloads the gamepack file, saving it in the {@link LynxConstants#SAVE_DIRECTORY}.
	 * 
	 * @param url The {@link URL} to download from.
	 * @param directory The directory to store the {@code gamepack} in.
	 * @return The path to the {@code gamepack} file.
	 * @throws IOException If there is an error downloading the {@code gamepack} file.
	 */
	private Path downloadGamepack(URL url, Path directory, String gpName) throws IOException {
		//gpName = gpName.replaceAll("\\\\", "\\\\\\\\");
		gpName = gpName.replace('/', '_').replace('*', '_');
		Path gamepack = directory.resolve(gpName);

		try (InputStream is = new BufferedInputStream(url.openStream());
				OutputStream os = new BufferedOutputStream(new FileOutputStream(gamepack.toFile()))) {

			int read;
			while ((read = is.read()) != -1) {
				os.write(read);
			}

		} catch (IOException e) {
			System.err.println("Error retrieving gamepack - please report along with the below stack trace:");
			throw new IOException(e);
		}

		return gamepack;
	}

	/**
	 * Gets the connection key (a 32-character string) from the {@link Map} of parameters.
	 * 
	 * @param parameters The map of parameter names to values.
	 * @return The key.
	 * @throws IllegalStateException If the map of parameters does not contain the connection key.
	 */
	private String getConnectionKey(Map<String, String> parameters) {
		for (String value : parameters.values()) {
			if (value.length() == 32) {
				return value;
			}
		}
		throw new IllegalStateException("Parameters did not contain the connection key - please report.");
	}

	/**
	 * Attempts to identify the current client version.
	 * 
	 * @param key The 32-character key.
	 * @return The client version.
	 * @throws IOException If there was an error identifying the version.
	 */
	private int identifyVersion(String key) throws IOException {
		try (ClientVersionWorker worker = new ClientVersionWorker(Js5Constants.HOST, key)) {
			worker.connect(Js5Constants.MAJOR_VERSION, Js5Constants.MINOR_VERSION);
			return worker.identifyVersion();
		} catch (IOException e) {
			System.err.println("Error identifying the correct version - please report:");
			throw new IOException(e);
		}
	}

	/**
	 * Initialises the retriever.
	 * 
	 * @throws IOException If there is an error creating the directory.
	 */
	public void init() throws IOException {
		Files.createDirectories(LynxConstants.SAVE_DIRECTORY);
	}

	/**
	 * Runs lynx, which downloads the gamepack, decrypts the {@code inner.pack.gz} file, and writes the class data.
	 * 
	 * @throws IOException If there is an I/O error.
	 */
	public String run() throws IOException {
		//long start = System.currentTimeMillis();

		Crawler crawler = new Crawler(new URL(LynxConstants.APPLET_URL + ",j0"));
		//ConfigCrawler crawler = new ConfigCrawler();
		Map<String, String> parameters = crawler.fetchParameters();

		//System.out.println("Successfully fetched parameters.");

		int version = 0;
		if (identifyVersion) {
			String key = getConnectionKey(parameters);
			version = identifyVersion(key);
		}

		String suffix = (version == 0 ? "local" : Integer.toString(version));
		Path directory = LynxConstants.SAVE_DIRECTORY.resolve(suffix);
		Files.createDirectories(directory);

		//System.out.println("Downloading gamepack: " + parameters.get("gamepack"));

		URL url = new URL(LynxConstants.APPLET_URL.replaceAll("g=runescape/", "") + parameters.get("gamepack"));
		Path gamepack = downloadGamepack(url, directory, parameters.get("gamepack"));

		String secret = parameters.get(LynxConstants.SECRET_PARAMETER_NAME);
		String vector = parameters.get(LynxConstants.VECTOR_PARAMETER_NAME);

		Map<String, ByteBuffer> classes;

		// Decrypt and decompress the inner pack.
		try (InnerPackDecrypter decrypter = new InnerPackDecrypter(gamepack, secret, vector)) {
			classes = decrypter.decrypt();
		} catch (GeneralSecurityException e) {
			System.out.println("Error decrypting the inner archive - please report:");
			throw new IllegalStateException(e);
		}

		//Path client = directory.resolve("bin");
		//Files.createDirectories(client);

		//System.out.println("Writing class files.");

		writeJar(classes, directory, "client.jar");
		//writeClasses(classes, client);
		writeParameters(parameters, directory, "parameters.txt");

		//System.out.println("Done, took " + (System.currentTimeMillis() - start) / 1000 + " seconds.");
		return suffix;
	}
	
	public void localDecrypt(String secret, String vector) throws Exception {
		InnerPackDecrypter decrypt = new InnerPackDecrypter(Paths.get(".", "repository/local").resolve("gamepack.jar"), secret, vector);
		Map<String, ByteBuffer> classes = decrypt.decrypt();
		decrypt.close();
		writeJar(classes, Paths.get(".", "repository/local"), "client.jar");
	}

	/**
	 * Writes the classes to the specified directory.
	 * 
	 * @param classes The {@link Map} of class names to {@link ByteBuffer}s.
	 * @param directory The {@link Path} to the directory to store the classes in.
	 * @throws IOException If there is an error writing the classes to files.
	 */
	@SuppressWarnings("unused")
	private void writeClasses(Map<String, ByteBuffer> classes, Path directory) throws IOException {
		for (Entry<String, ByteBuffer> entry : classes.entrySet()) {
			String name = entry.getKey();
			Path clazz = directory.resolve(name);

			if (name.contains("/")) {
				Files.createDirectories(clazz.getParent());
			}

			try (FileOutputStream os = new FileOutputStream(clazz.toFile()); FileChannel channel = os.getChannel()) {
				ByteBuffer buffer = entry.getValue();
				channel.write(buffer);
			} catch (IOException e) {
				System.err.println("Error writing classes to file - please ensure this program has write permissions.");
				throw new IOException(e);
			}
		}
	}

	/**
	 * Writes the class data to a jar file with the specified name.
	 * 
	 * @param classes The {@link Map} of class names to {@link ByteBuffer}s.
	 * @param directory The {@link Path} to the directory to store the jar file in.
	 * @param name The name of the jar file.
	 * @throws IOException If there is an error writing to the jar file.
	 */
	private void writeJar(Map<String, ByteBuffer> classes, Path directory, String name) throws IOException {
		Path jar = directory.resolve(name);

		try (JarOutputStream jos = new JarOutputStream(new BufferedOutputStream(new FileOutputStream(jar.toFile())))) {
			for (Entry<String, ByteBuffer> clazz : classes.entrySet()) {
				ZipEntry zip = new ZipEntry(clazz.getKey());

				jos.putNextEntry(zip);
				jos.write(clazz.getValue().array());
			}
		} catch (IOException e) {
			System.err.println("Error writing classes to jar - please ensure this program has write permissions.");
			throw new IOException(e);
		}
	}
	
	/**
	 * Writes the parameters to the specified name.
	 * 
	 * @param parameters The {@link Map} of keys and values
	 * @param directory The {@link Path} to the directory to store the jar file in.
	 * @param name The name of the txt file.
	 * @throws IOException If there is an error writing to the jar file.
	 */
	private void writeParameters(Map<String, String> params, Path directory, String name) throws IOException {
		Path path = directory.resolve(name);
		try {
			PrintWriter out = new PrintWriter(path.toFile());
			for (Entry<String, String> entry : params.entrySet()) {
				if (entry.getKey().startsWith("b")
						|| entry.getKey().startsWith("c")
						|| entry.getKey().startsWith("g")
						|| entry.getKey().startsWith("h")
						|| entry.getKey().startsWith("i")
						|| entry.getKey().startsWith("j")
						|| entry.getKey().startsWith("s"))
					continue;
				
				out.println("client_parameters.put(\"" + entry.getKey() + "\", \"" + entry.getValue() + "\");");
			}
			out.close();
		} catch (IOException e) {
			System.err.println("Error writing parameters to txt - please ensure this program has write permissions.");
			throw new IOException(e);
		}
	}

}